==> ben-2.16.tar.gz: ben/ini.c <==
/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <string.h>
#include "xmem.h"
#include "ini.h"

// **************************************************************************
// 
// **************************************************************************
static int ini_needs_quote(char *str)
{
	size_t l = 0;
	
	while (1) {
		char c = str[l];
		
		if (c == 0) {
			// empty string needs to be quoted
			return l == 0;
		}
		
		// space, tabs, newlines would end a token
		// an initial quote would indicate a quoted string
		if ((c == ' ') || (c == '\t') || (c == '\n') || (c == '"')) {
			return 1;
		}
		
		l++;
	}
}

static void ini_append_value(buffer *b, char *str)
{
	if (!ini_needs_quote(str)) {
		buffer_strappend(b, str);
		return;
	}

	buffer_memappend(b, "\"", 1);
	
	char *s = str;
	
	while (1) {
		char c = *s;
		
		if (c == 0)
			break;
		
		if (c == '\\') {
			buffer_memappend(b, "\\\\", 2);
		} else if (c == '"') {
			buffer_memappend(b, "\\\"", 2);
		} else if (c == '\n') {
			buffer_memappend(b, "\\n", 2);
		} else {
			buffer_memappend(b, &c, 1);
		}
		
		s++;
	}

	buffer_memappend(b, "\"", 1);
}

// **************************************************************************
// 
// **************************************************************************
void ini_section(buffer *b, char *name)
{
	buffer_memappend(b, "[", 1);
	buffer_strappend(b, name);
	buffer_memappend(b, "]\n", 2);
}

void ini_entry(buffer *b, char *name, char *value)
{
	buffer_memappend(b, "\t", 1);
	buffer_strappend(b, name);
	buffer_memappend(b, " = ", 3);
	ini_append_value(b, value);
	buffer_memappend(b, "\n", 1);
}

void ini_int(buffer *b, char *name, int value)
{
	char str[32];
	
	xsnprintf(str, sizeof(str), "%d", value);
	
	ini_entry(b, name, str);
}

void ini_date(buffer *b, char *name, time_t value)
{
	char str[20];
	
	date_str(str, sizeof(str), value);
	
	ini_entry(b, name, str);
}

// **************************************************************************
// 
// **************************************************************************
#define INI_WHILE	0
#define INI_UNTIL	1
static int ini_loop(buffer *b, int type, char *list)
{
	char *s = b->data;
	size_t size = b->size;
	size_t o = b->partial;
	
	while (1) {
		if (o >= size) {
			b->partial = o;
			return -1;
		}
		
		char c = s[o];
		
		int found = (strchr(list, c) != NULL);
		
		if (((type == INI_WHILE) && (!found))
		||  ((type == INI_UNTIL) && (found))) {
			b->partial = o;
			return 0;
		}
		
		o++;
	}
}

static int ini_skip(buffer *b, char *list)
{
	return ini_loop(b, INI_WHILE, list);
}

static int ini_seek(buffer *b, char *list)
{
	return ini_loop(b, INI_UNTIL, list);
}

static char *ini_cur(buffer *b)
{
	char *s = b->data;
	return &s[b->partial];
}

static int ini_pop(buffer *b, char ref)
{
	if (b->partial >= b->size)
		return 0;
	
	if (*ini_cur(b) != ref)
		return 0;
	
	b->partial++;
	return 1;
}

static int ini_unquote(buffer *b)
{
	char *s = b->data;
	size_t size = b->size;
	size_t o = b->partial;
	size_t t = o;
	
	while (1) {
		if (o >= size) {
			b->partial = o;
			return -1;
		}
		
		char c = s[o];
		
		if (c == '\\') {
			c = s[o + 1];
			
			if ((c == '\\') || (c == '"')) {
				s[t] = c;
				t++;
			} else if (c == 'n') {
				s[t] = '\n';
				t++;
			} else if (c == '\n') {
			} else {
				s[t + 0] = '\\';
				s[t + 1] = c;
				t += 2;
			}
			
			o += 2;
		} else if (c == '"') {
			s[t] = 0;
			b->partial = o + 1;
			return 0;
		} else {
			s[t] = c;
			t++;
			o++;
		}
	}
}

// **************************************************************************
// 
// **************************************************************************
int ini_get(buffer *b, char **name_r, char **value_r)
{
	// return null on eof or error
	if (name_r)
		*name_r = NULL;
	if (value_r)
		*value_r = NULL;
	
	// initial blank space, may encounter eof
	if (ini_skip(b, " \t\n"))
		return 0;
	
	// section
	if (ini_pop(b, '[')) {
		if (ini_skip(b, " \t\n")) {
			fprintf(stderr, "ini: eof in section name\n");
			return -1;
		}
		
		char *section_start = ini_cur(b);
		
		if (ini_seek(b, " \t\n[]")) {
			fprintf(stderr, "ini: eof in section name\n");
			return -1;
		}
		
		char *section_stop = ini_cur(b);
		
		if (ini_skip(b, " \t\n")) {
			fprintf(stderr, "ini: eof in section name\n");
			return -1;
		}
		
		if (!ini_pop(b, ']')) {
			fprintf(stderr, "ini: no closing square bracket\n");
			return -1;
		}
		
		*section_stop = 0;
		if (name_r)
			*name_r = section_start;
		
		return 0;
	}
	
	// assignment name
	char *name_start = ini_cur(b);
	
	if (ini_seek(b, " \t\n=")) {
		fprintf(stderr, "ini: missing = sign\n");
		return -1;
	}
	
	char *name_end = ini_cur(b);
	
	if (ini_skip(b, " \t\n")) {
		fprintf(stderr, "ini: missing = sign\n");
		return -1;
	}
	
	if (!ini_pop(b, '=')) {
		fprintf(stderr, "ini: expected = sign\n");
		return -1;
	}
	
	if (ini_skip(b, " \t\n")) {
		fprintf(stderr, "ini: expected value after = sign\n");
		return -1;
	}
	
	// assignment value
	char *val_start;
	
	if (ini_pop(b, '"')) {
		val_start = ini_cur(b);
		
		if (ini_unquote(b)) {
			fprintf(stderr, "ini: unended quoted string\n");
			return -1;
		}
	} else {
		val_start = ini_cur(b);
		
		if (ini_seek(b, " \t\n")) {
			fprintf(stderr, "ini: empty unquoted value\n");
			return -1;
		}
		
		*(ini_cur(b)) = 0;
	}
	
	*name_end = 0;
	
	if (name_r)
		*name_r = name_start;
	if (value_r)
		*value_r = val_start;
	
	return 0;
}


// **************************************************************************
// 
// **************************************************************************

==> ben-2.17.tar.gz: ben/ini.c <==
/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <string.h>
#include "xmem.h"
#include "ini.h"

// **************************************************************************
// 
// **************************************************************************
static int ini_needs_quote(char *str)
{
	size_t l = 0;
	
	while (1) {
		char c = str[l];
		
		if (c == 0) {
			// empty string needs to be quoted
			return l == 0;
		}
		
		// space, tabs, newlines would end a token
		// an initial quote would indicate a quoted string
		if ((c == ' ') || (c == '\t') || (c == '\n') || (c == '"')) {
			return 1;
		}
		
		l++;
	}
}

static void ini_append_value(buffer *b, char *str)
{
	if (!ini_needs_quote(str)) {
		buffer_strappend(b, str);
		return;
	}

	buffer_memappend(b, "\"", 1);
	
	char *s = str;
	
	while (1) {
		char c = *s;
		
		if (c == 0)
			break;
		
		if (c == '\\') {
			buffer_memappend(b, "\\\\", 2);
		} else if (c == '"') {
			buffer_memappend(b, "\\\"", 2);
		} else if (c == '\n') {
			buffer_memappend(b, "\\n", 2);
		} else {
			buffer_memappend(b, &c, 1);
		}
		
		s++;
	}

	buffer_memappend(b, "\"", 1);
}

// **************************************************************************
// 
// **************************************************************************
void ini_section(buffer *b, char *name)
{
	buffer_memappend(b, "[", 1);
	buffer_strappend(b, name);
	buffer_memappend(b, "]\n", 2);
}

void ini_entry(buffer *b, char *name, char *value)
{
	buffer_memappend(b, "\t", 1);
	buffer_strappend(b, name);
	buffer_memappend(b, " = ", 3);
	ini_append_value(b, value);
	buffer_memappend(b, "\n", 1);
}

void ini_int(buffer *b, char *name, int value)
{
	char str[32];
	
	xsnprintf(str, sizeof(str), "%d", value);
	
	ini_entry(b, name, str);
}

void ini_duration(buffer *b, char *name, gtime value)
{
	ini_entry(b, name, duration_str(value, GTIME_DAYS | GTIME_SUBSEC).s);
}

void ini_date(buffer *b, char *name, gtime value)
{
	ini_entry(b, name, date_str(value, GTIME_SUBSEC).s);
}

// **************************************************************************
// 
// **************************************************************************
#define INI_WHILE	0
#define INI_UNTIL	1
static int ini_loop(buffer *b, int type, char *list)
{
	char *s = b->data;
	size_t size = b->size;
	size_t o = b->partial;
	
	while (1) {
		if (o >= size) {
			b->partial = o;
			return -1;
		}
		
		char c = s[o];
		
		int found = (strchr(list, c) != NULL);
		
		if (((type == INI_WHILE) && (!found))
		||  ((type == INI_UNTIL) && (found))) {
			b->partial = o;
			return 0;
		}
		
		o++;
	}
}

static int ini_skip(buffer *b, char *list)
{
	return ini_loop(b, INI_WHILE, list);
}

static int ini_seek(buffer *b, char *list)
{
	return ini_loop(b, INI_UNTIL, list);
}

static char *ini_cur(buffer *b)
{
	char *s = b->data;
	return &s[b->partial];
}

static int ini_pop(buffer *b, char ref)
{
	if (b->partial >= b->size)
		return 0;
	
	if (*ini_cur(b) != ref)
		return 0;
	
	b->partial++;
	return 1;
}

static int ini_unquote(buffer *b)
{
	char *s = b->data;
	size_t size = b->size;
	size_t o = b->partial;
	size_t t = o;
	
	while (1) {
		if (o >= size) {
			b->partial = o;
			return -1;
		}
		
		char c = s[o];
		
		if (c == '\\') {
			c = s[o + 1];
			
			if ((c == '\\') || (c == '"')) {
				s[t] = c;
				t++;
			} else if (c == 'n') {
				s[t] = '\n';
				t++;
			} else if (c == '\n') {
			} else {
				s[t + 0] = '\\';
				s[t + 1] = c;
				t += 2;
			}
			
			o += 2;
		} else if (c == '"') {
			s[t] = 0;
			b->partial = o + 1;
			return 0;
		} else {
			s[t] = c;
			t++;
			o++;
		}
	}
}

// **************************************************************************
// 
// **************************************************************************
int ini_get(buffer *b, char **name_r, char **value_r)
{
	// return null on eof or error
	if (name_r)
		*name_r = NULL;
	if (value_r)
		*value_r = NULL;
	
	// initial blank space, may encounter eof
	if (ini_skip(b, " \t\n"))
		return 0;
	
	// section
	if (ini_pop(b, '[')) {
		if (ini_skip(b, " \t\n")) {
			fprintf(stderr, "ini: eof in section name\n");
			return -1;
		}
		
		char *section_start = ini_cur(b);
		
		if (ini_seek(b, " \t\n[]")) {
			fprintf(stderr, "ini: eof in section name\n");
			return -1;
		}
		
		char *section_stop = ini_cur(b);
		
		if (ini_skip(b, " \t\n")) {
			fprintf(stderr, "ini: eof in section name\n");
			return -1;
		}
		
		if (!ini_pop(b, ']')) {
			fprintf(stderr, "ini: no closing square bracket\n");
			return -1;
		}
		
		*section_stop = 0;
		if (name_r)
			*name_r = section_start;
		
		return 0;
	}
	
	// assignment name
	char *name_start = ini_cur(b);
	
	if (ini_seek(b, " \t\n=")) {
		fprintf(stderr, "ini: missing = sign\n");
		return -1;
	}
	
	char *name_end = ini_cur(b);
	
	if (ini_skip(b, " \t\n")) {
		fprintf(stderr, "ini: missing = sign\n");
		return -1;
	}
	
	if (!ini_pop(b, '=')) {
		fprintf(stderr, "ini: expected = sign\n");
		return -1;
	}
	
	if (ini_skip(b, " \t\n")) {
		fprintf(stderr, "ini: expected value after = sign\n");
		return -1;
	}
	
	// assignment value
	char *val_start;
	
	if (ini_pop(b, '"')) {
		val_start = ini_cur(b);
		
		if (ini_unquote(b)) {
			fprintf(stderr, "ini: unended quoted string\n");
			return -1;
		}
	} else {
		val_start = ini_cur(b);
		
		if (ini_seek(b, " \t\n")) {
			fprintf(stderr, "ini: empty unquoted value\n");
			return -1;
		}
		
		*(ini_cur(b)) = 0;
	}
	
	*name_end = 0;
	
	if (name_r)
		*name_r = name_start;
	if (value_r)
		*value_r = val_start;
	
	return 0;
}


// **************************************************************************
// 
// **************************************************************************
